"""Repository interfaces for NWP data sources and stores.

These interfaces define the signatures that *driven* actors must conform to
in order to interact with the core.
Also sometimes referred to as *secondary ports*.

All NWP providers use some kind of model to generate their data. This model
can be physics-based, such as ERA5, or a machine learning model, such as
Google's GraphCast. The `ModelMetadata` object is used to abstract the
differences between these models, allowing the core to interact with them
in a uniform way, via the `RawRepository` interface.
"""

import abc
import datetime as dt
import logging
from collections.abc import Callable, Iterator

import xarray as xr
from returns.result import ResultE

from nwp_consumer.internal import entities

log = logging.getLogger("nwp-consumer")


class RawRepository(abc.ABC):
    """Interface for a repository that produces raw NWP data.

    Since different producers of NWP data have different data storage
    implementations, a RawRepository needs to define its own download
    and processing methods.

    A source may provide one or more files for a given init time.
    To keep memory usage at a minimum, when converting raw data to zarr,
    converted data is persisted to disk in a store.
    In this manner, writes can be done in parallel, but a schema needs to be known
    in advance.

    As such, an important distinction is made between:
        - the *fileset*: Raw store data for an init time
        - the *store*: The Zarr store containing the processed data
    """

    @classmethod
    @abc.abstractmethod
    def authenticate(cls) -> ResultE["RawRepository"]:
        """Create a new authenticated instance of the class."""
        pass

    @abc.abstractmethod
    def fetch_init_data(
        self, it: dt.datetime,
    ) -> Iterator[Callable[..., ResultE[list[xr.DataArray]]]]:
        """Fetch raw data files for an init time as xarray datasets.

        As per the typing, the return value is a generator of functions that
        may produce one or more xarray datasets.
        The generator-of-functions approach (typed here as ``Iterator[Callable...]``)
        is important, as it allows for lazy evaluation:
        by returning a generator of delayed objects, joblib can parallelize
        the download and the results can be accumulated in a low-memory fashion (see
        `the JobLib documentation on parallel generators
        <https://joblib.readthedocs.io/en/stable/auto_examples/parallel_generator.html>`_).

        An example psuedocode implementation is shown below:

        >>> from joblib import delayed
        >>> from returns.result import Result, ResultE
        >>> from typing import override
        >>> from collections.abc import Callable, Iterator
        >>> import xarray as xr
        >>> import datetime as dt
        >>>
        >>> # Pseudocode for a raw repository
        >>> class MyRawRepository(RawRepository):
        ...     @override
        ...     def fetch_init_data(self, it: dt.datetime) \
        ...             -> Iterator[Callable[..., ResultE[list[xr.DataArray]]]]:
        ...         for file in ["raw_file1.grib", "raw_file2.grib"]:
        ...             yield delayed(self._download_and_convert)(file)
        ...
        ...     def _download_and_convert(self, file: str) -> ResultE[list[xr.DataArray]]:
        ...         '''Download and convert a raw file to an xarray dataset.'''
        ...         return Success([xr.open_dataset(file).to_dataarray()])

        .. warning:: No downloading or processing should be done in this method*. All of that
          should be handled in the function that is yielded by the generator -
          ``_download_and_convert`` in the example above.
          This is to allow for parallelization of the download and processing.

        .. note:: It is however, worth considering the most efficient way to download and process
          the data. The above assumes that the data comes in many files, but there is a possibility
          of the case where the source provides one large file with many underlying datasets within.
          In this case, it may be more efficient to download the large file in the
          `fetch_init_data` method and then process the datasets within via the yielded functions.

        .. note:: For the moment, this returns a list of ``xarray.DataArray`` objects. It may be
          more efficient to return a generator here to avoid reading all the datasets into
          memory at once, however, often the source of these datasets is ``cfgrib.open_datasets``
          which has no option for returning a generator, hence the current choice of ``list``.
          This may be revisited in the future, for instance by recreating the ``open_datasets``
          function in a manner which returns a generator of datasets.

        Args:
            it: The initialization time for which to fetch data.

        Returns:
            A generator of delayed xarray dataarrays for the init time.
        """
        pass

    @staticmethod
    @abc.abstractmethod
    def repository() -> entities.RawRepositoryMetadata:
        """Metadata about the model repository."""
        pass

    @staticmethod
    @abc.abstractmethod
    def model() -> entities.ModelMetadata:
        """Metadata about the model."""
        pass


class NotificationRepository(abc.ABC):
    """Interface for a repository that sends notifications.

    Adaptors for this port enable sending notifications to
    a desired notification_repositories channel.
    """

    @abc.abstractmethod
    def notify(
        self,
        message: entities.StoreAppendedNotification | entities.StoreCreatedNotification,
    ) -> ResultE[str]:
        """Send a notification_repositories."""
        pass



"""NWP Consumer.

Usage Documentation
===================

Configuration
-------------

The following environment variables can be used to configure the application:

.. code-block:: none

    | Key                       | Description                         | Default                                     |
    |---------------------------|-------------------------------------|---------------------------------------------|
    | LOGLEVEL                  | The logging level for the app.      | INFO                                        |
    |---------------------------|-------------------------------------|---------------------------------------------|
    | RAWDIR                    | The working directory for the app.  | ~/.local/cache/nwp/<REPO>/<MODEL>/raw       |
    |                           | Can be a local path or an S3 URI.   |                                             |
    |---------------------------|-------------------------------------|---------------------------------------------|
    | ZARRDIR                   | The output directory for the app.   | ~/.local/cache/nwp/<REPO>/<MODEL>/data      |
    |                           | Can be a local path or an S3 URI.   |                                             |
    |---------------------------|-------------------------------------|---------------------------------------------|
    | NOTIFICATION_REPOSITORY   | The notification repository to use. | stdout                                      |
    |---------------------------|-------------------------------------|---------------------------------------------|
    | MODEL_REPOSITORY          | The model repository to use.        | ceda-metoffice-global                       |
    |---------------------------|-------------------------------------|---------------------------------------------|
    | CONCURRENCY               | Whether to use concurrency.         | True                                        |
    |---------------------------|-------------------------------------|---------------------------------------------|

There is also specific configuration variables for some model repositories.
Refer to their documentation for more information: `nwp_consumer.internal.repositories`.


Development Documentation
=========================

Getting started for development
-------------------------------

In order to work on the project, first clone the repository.
Then, create a virtual environment and install the dependencies
using an editable pip installation::

    $ git clone git@github.com:openclimatefix/nwp-consumer.git
    $ cd nwp-consumer
    $ python -m venv ./venv
    $ source ./venv/bin/activate
    $ pip install -e .[dev]

.. note:: ZSH users may have to escape the square brackets in the last command.

This enables the use of the 'nwp-consumer-cli' command in the virtualenv, which
runs the `nwp_consumer.cmd.main.run_cli` entrypoint. The editable installation
ensures that changes to the code are immediately reflected while using the command.


Project structure
-----------------

The code is structured following principles from the `Hexagonal Architecture`_ pattern.
In brief, this means a clear separation between
the application's business logic - it's *core* - and the *actors* that are external to it.

The core of the services is split into three main components:

- `nwp_consumer.internal.entities` - The domain classes that define the structure of the data
  that the services works with, and the business logic they contain.
- `nwp_consumer.internal.ports` - The interfaces that define how the services interact with external actors.
- `nwp_consumer.internal.services` - The business logic that defines how the service functions.

Alongside these core components are the actors, which adhere to the interfaces defined in the
ports module. Actors come in two flavours, *driving* and *driven*.
Driven actors are sources and sinks of data, such as databases and message queues,
while driving actors are methods of interacting with the core, such as a command-line interface
or REST server.

This application currently has the following defined actors:

- `nwp_consumer.internal.repositories.raw_repositories` (driven) - The sources of NWP data.
- `nwp_consumer.internal.repositories.notification_repositories` (driven) - The sinks of notification data.
- `nwp_consumer.internal.handlers.cli` (driving) - The command-line interface for the services.

The actors are then responsible for implementing the abstract ports,
and are *dependency-injected* in at runtime. This allows the services to be easily tested
and extended. See 'further reading' for more information.

Head into `nwp_consumer.internal` to see the details of each of these components.

Where do I go to...?
--------------------

- **...modify the business logic?** Check out the `internal.services` module.
- **...add a new source of NWP data?** Implement a new repository in `internal.repositories.raw_repositories`.
- **...modify the command line interface?** Check out `internal.handlers.cli`.

Further reading
===============

On packaging a python project using setuptools and pyproject.toml:

- The official PyPA packaging guide: https://packaging.python.org/
- A step-by-step practical guide on the *godatadriven* blog:
  https://godatadriven.com/blog/a-practical-guide-to-setuptools-and-pyproject-toml/
- The pyproject.toml metadata specification:
  https://packaging.python.org/en/latest/specifications/declaring-project-metadata

On hexagonal architecture:

- A concrete example using Python:
  https://medium.com/towards-data-engineering/a-concrete-example-of-the-hexagonal-architecture-in-python-d821213c6fb9
- An overview of the fundamentals incorporating Typescript:
  https://medium.com/ssense-tech/hexagonal-architecture-there-are-always-two-sides-to-every-story-bc0780ed7d9c

- Another example using Go:
  https://medium.com/@matiasvarela/hexagonal-architecture-in-go-cfd4e436faa3

On the directory structure:

- The official PyPA discussion on src and flat layouts"
  https://packaging.python.org/en/latest/discussions/src-layout-vs-flat-layout/

.. _Hexagonal Architecture: https://alistair.cockburn.us/hexagonal-architecture/
"""

import logging
import sys
import os

if sys.stdout.isatty():
    # Simple logging for terminals
    _formatstr="%(levelname)s [%(name)s] | %(message)s"
else:
    # JSON logging for containers
    _formatstr="".join((
        "{",
        '"message": "%(message)s", ',
        '"severity": "%(levelname)s", "timestamp": "%(asctime)s.%(msecs)03dZ", ',
        '"logging.googleapis.com/labels": {"python_logger": "%(name)s"}, ',
        '"logging.googleapis.com/sourceLocation": ',
        '{"file": "%(filename)s", "line": %(lineno)d, "function": "%(funcName)s"}',
        "}",
    ))

_loglevel: int | str = logging.getLevelName(os.getenv("LOGLEVEL", "INFO").upper())
logging.basicConfig(
    level=logging.INFO if isinstance(_loglevel, str) else _loglevel,
    stream=sys.stdout,
    format=_formatstr,
    datefmt="%Y-%m-%dT%H:%M:%S",
)

for logger in [
    "numcodecs",
    "numexpr",
    "gribapi",
    "aiobotocore",
    "s3fs",
    "fsspec",
    "asyncio",
    "botocore",
    "cfgrib",
]:
    logging.getLogger(logger).setLevel(logging.WARNING)



"""Tests for the ModelMetadata class."""

import dataclasses
import unittest

from hypothesis import given
from hypothesis import strategies as st

from .coordinates import NWPDimensionCoordinateMap
from .modelmetadata import ModelMetadata, Models
from .parameters import Parameter


class TestModelMetadata(unittest.TestCase):
    """Tests for the ModelMetadata class."""

    def test_with_region(self) -> None:
        """Test the with_region method."""
        metadata: ModelMetadata = ModelMetadata(
            name="test",
            resolution="0.1 degrees",
            expected_coordinates=NWPDimensionCoordinateMap(
                init_time=[],
                step=list(range(0, 85, 1)),
                variable=[Parameter.WIND_U_COMPONENT_10m],
                latitude=[float(f"{lat / 10:.2f}") for lat in range(900, -900 - 1, -1)],
                longitude=[float(f"{lon / 10:.2f}") for lon in range(-1800, 1800 + 1, 1)],
            ),
            running_hours=[0, 6, 12, 18],
            delay_minutes=60,
        )

        @dataclasses.dataclass
        class TestCase:
            region: str
            expected_name: str
            expected_lat: list[float]
            expected_lon: list[float]

        test_cases: list[TestCase] = [
            TestCase(
                region="uk",
                expected_name="test_uk",
                expected_lat=[float(f"{lat / 10:.2f}") for lat in range(620, 480 - 1, -1)],
                expected_lon=[float(f"{lon / 10:.2f}") for lon in range(-120, 30 + 1, 1)],
            ),
            TestCase(
                region="india",
                expected_name="test_india",
                expected_lat=[float(f"{lat / 10:.2f}") for lat in range(350, 60 - 1, -1)],
                expected_lon=[float(f"{lon / 10:.2f}") for lon in range(670, 970 + 1, 1)],
            ),
            TestCase(
                region="invalid",
                expected_name="test",
                expected_lat=[float(f"{lat / 10:.2f}") for lat in range(900, -900 - 1, -1)],
                expected_lon=[float(f"{lon / 10:.2f}") for lon in range(-1800, 1800 + 1, 1)],
            ),
        ]

        for test_case in test_cases:
            new_metadata = metadata.with_region(test_case.region)
            self.assertEqual(new_metadata.name, test_case.expected_name)
            self.assertListEqual(
                new_metadata.expected_coordinates.latitude,  # type: ignore
                test_case.expected_lat,
            )
            self.assertListEqual(
                new_metadata.expected_coordinates.longitude,  # type: ignore
                test_case.expected_lon,
            )

    @given(
        st.sampled_from([attr for attr in dir(Models) if not attr.startswith("__")]),
        st.sampled_from(["uk", "india", "invalid"]),
    )
    def hypothesis_test_with_region(self, model: ModelMetadata, region: str) -> None:
        """Assert that with_region does not raise an exception."""
        model.with_region(region)



"""Implementation of the NWP consumer service."""

import dataclasses
import datetime as dt
import functools
import logging
import os
import pathlib
from collections.abc import Callable, Iterator
from typing import override

import xarray as xr
from joblib import Parallel, cpu_count
from returns.methods import partition
from returns.pipeline import flow
from returns.result import Failure, ResultE, Success

from nwp_consumer.internal import entities, ports

log = logging.getLogger("nwp-consumer")


class ConsumerService(ports.ConsumeUseCase):
    """Service implementation for the NWP Consumer.

    Defines the business-critical methods and logic.
    """

    mr: ports.RawRepository
    nr: ports.NotificationRepository

    def __init__(
        self,
        model_repository: ports.RawRepository,
        notification_repository: ports.NotificationRepository,
    ) -> None:
        """Create a new instance of the service."""
        self.mr = model_repository
        self.nr = notification_repository

    @classmethod
    def from_adaptors(
        cls,
        model_adaptor: type[ports.RawRepository],
        notification_adaptor: type[ports.NotificationRepository],
    ) -> ResultE["ConsumerService"]:
        """Create a new instance of the service from adaptors."""
        notification_repository = notification_adaptor()
        model_repository_result = model_adaptor.authenticate()
        return model_repository_result.do(
            cls(
                model_repository=model_repository,
                notification_repository=notification_repository,
            )
            for model_repository in model_repository_result
        )

    @staticmethod
    def _fold_dataarrays_generator(
        generator: Iterator[ResultE[list[xr.DataArray]]],
        store: entities.TensorStore,
    ) -> ResultE[int]:
        """Process data from data generator.

        Args:
            generator: A generator of ResultE objects containing either a list data arrays
                or a Failure object.
            store: The store to write the data to.

        Returns:
            A ResultE object containing the sum of the write results or a Failure object.
        """
        results: list[ResultE[int]] = []
        for value in generator:
            if isinstance(value, Failure):
                results.extend([value])
            elif len(value.unwrap()) == 0:
                continue
            else:
                results.extend([store.write_to_region(da=da) for da in value.unwrap()])
        successes, failures = partition(results)
        # TODO: Define the failure threshold for number of write attempts properly
        # ECMWF Realtime for instance needs a bit of tolerance because of some files not containing
        # data for all geographic regions.
        log.info(f"Processed {len(successes)} DataArrays successfully with {len(failures)} errors.")
        if len(failures) / len(results) > 0.06:
            for i, exc in enumerate(failures):
                if i < 5:
                    log.error(str(exc))
                else:
                    break
            return Failure(
                OSError(
                    "Error threshold exceeded: "
                    f"{len(failures)/len(results)} errors (>6%) occurred during processing.",
                ),
            )
        else:
            return Success(sum(successes))

    @staticmethod
    def _parallelize_generator[T](
        delayed_generator: Iterator[Callable[..., T]],
        max_connections: int,
    ) -> Iterator[T]:
        """Parallelize a generator of delayed functions.

        Args:
            delayed_generator: An iterable of delayed items.
                The creation of these items must be delayed, either via joblib.delayed
                or functools.partial, so they can be executed lazily.
            max_connections: The maximum number of connections to use.
        """
        # TODO: Change this based on threads instead of CPU count
        # TODO: Enable choosing between threads and processes?
        n_jobs: int = max(cpu_count() - 1, max_connections)
        prefer = "threads"

        if os.getenv("CONCURRENCY", "True").capitalize() == "False":
            n_jobs = 1

        log.debug(f"Using {n_jobs} concurrent {prefer}")

        return Parallel(  # type: ignore
            n_jobs=n_jobs,
            prefer=prefer,
            verbose=0,
            return_as="generator_unordered",
        )(delayed_generator)

    @staticmethod
    def _create_suitable_store(
        repository_metadata: entities.RawRepositoryMetadata,
        model_metadata: entities.ModelMetadata,
        period: dt.datetime | dt.date | None = None,
    ) -> ResultE[entities.TensorStore]:
        """Create a store for the data with the relevant init time coordinates.

        Args:
            repository_metadata: The metadata for the repository.
            model_metadata: The metadata for the model.
            period: The period for which to gather init time data.
        """
        its: list[dt.datetime] = []
        match period:
            case _ if period is None:
                its = [
                    repository_metadata.determine_latest_it_from(
                        t=dt.datetime.now(tz=dt.UTC),
                        running_hours=model_metadata.running_hours,
                        delay_minutes=model_metadata.delay_minutes,
                    ),
                ]
            case single_it if isinstance(period, dt.datetime):
                its = [single_it]  # type: ignore
            case multiple_its if isinstance(period, dt.date):
                its = model_metadata.month_its(
                    year=multiple_its.year,
                    month=multiple_its.month,
                )
        its = [it.replace(tzinfo=dt.UTC) for it in its]

        # Create a store for the data with the relevant init time coordinates
        return entities.TensorStore.initialize_empty_store(
            model=model_metadata.name,
            repository=repository_metadata.name,
            coords=dataclasses.replace(
                model_metadata.expected_coordinates,
                init_time=its,
            ),
            chunks=model_metadata.expected_coordinates.chunking(
                chunk_count_overrides=model_metadata.chunk_count_overrides,
            ),
        )

    @override
    def consume(
        self,
        period: dt.datetime | dt.date | None = None,
        delete_on_failure: bool = True,
    ) -> ResultE[str]:
        """Consume NWP data to Zarr format for desired time period.

        Where possible the implementation should be as memory-efficient as possible.
        The designs of the repository methods also enable parallel processing within
        the implementation.

        Args:
            period: The period for which to gather init time data.
            delete_on_failure: Whether to delete the store if the operation fails.

        Returns:
            The path to the produced Zarr store.

        See Also:
            - `repositories.RawRepository.fetch_init_data`
            - `tensorstore.TensorStore.write_to_region`
            - https://joblib.readthedocs.io/en/stable/auto_examples/parallel_generator.html
        """
        monitor = entities.PerformanceMonitor()
        with monitor:
            init_store_result = self._create_suitable_store(
                repository_metadata=self.mr.repository(),
                model_metadata=self.mr.model(),
                period=period,
            )
            if isinstance(init_store_result, Failure):
                return Failure(
                    OSError(
                        f"Failed to initialize store for init time: {init_store_result!s}",
                    ),
                )
            store = init_store_result.unwrap()

            missing_times_result = store.missing_times()
            if isinstance(missing_times_result, Failure):
                delete_store_result = store.delete_store()
                if isinstance(delete_store_result, Failure):
                    log.error(
                        f"Failed to delete store after error: {delete_store_result}",
                    )
                return missing_times_result

            for n, it in enumerate(missing_times_result.unwrap()):
                log.info(
                    f"Consuming data from {self.mr.repository().name} for {it:%Y-%m-%d %H:%M} "
                    f"(time {n + 1}/{len(missing_times_result.unwrap())})",
                )
                process_result = flow(
                    self._parallelize_generator(
                        self.mr.fetch_init_data(it),
                        max_connections=self.mr.repository().max_connections,
                    ),
                    functools.partial(self._fold_dataarrays_generator, store=store),
                )
                if isinstance(process_result, Failure):
                    delete_store_result = store.delete_store()
                    if isinstance(delete_store_result, Failure):
                        log.error(
                            f"Failed to delete store after error: {delete_store_result}",
                        )
                    return process_result

            validation_result = store.validate_store()
            if isinstance(validation_result, Failure) and delete_on_failure:
                delete_store_result = store.delete_store()
                if isinstance(delete_store_result, Failure):
                    log.error(
                        f"Failed to delete store after error: {delete_store_result}",
                    )
                return validation_result

        notification_message = entities.StoreCreatedNotification(
            filename=pathlib.Path(store.path).name,
            size_mb=store.size_kb // 1024,
            performance=entities.PerformanceMetadata(
                duration_seconds=monitor.get_runtime(),
                memory_mb=monitor.max_memory_mb(),
            ),
        )
        notify_result = self.nr.notify(message=notification_message)
        if isinstance(notify_result, Failure):
            log.error(
                "Failed to notify of store creation: "
                f"{notify_result.failure()}. "
                f"Notification: {notification_message}",
            )

        log.info(f"Successfully processed data to '{store.path}'")
        return Success(store.path)

    @override
    def archive(self, period: dt.date) -> ResultE[str]:
        return self.consume(period=period)

    @staticmethod
    def info(
        model_adaptor: type[ports.RawRepository],
        notification_adaptor: type[ports.NotificationRepository],
    ) -> str:
        """Get information about the service."""
        raise NotImplementedError("Not yet implemented")



"""Repository implementation for data from MetOffice's DataHub service.

Repository Information
======================

The API documentation for the MetOffice Weather Datahub can be found at:
https://datahub.metoffice.gov.uk/docs/f/category/atmospheric/type/atmospheric/api-documentation

Documented Structure
--------------------

MetOffice provide a number of models, a few of which OCF consume. Their flagship deterministic
model us called the "Unified Model" (UM) and is run in two configurations: "Global" and "UK".
The "Global" model has a resolution of 10km and the "UK" model has a resolution of ~2km.

See https://datahub.metoffice.gov.uk/docs/f/category/atmospheric/overview for more information.

Data is provided on a per-order basis, so the filestructure depends on the order ID.
For OCF's purposes, on file per parameter per step is requested.


Actual Structure
----------------

For the global model, the grid of the provided data matches the described structure.
The latitude and longitude increments are ascertained from the GRIB2 file's metadata:
.. code-block:: none

    iDirectionIncrementInDegrees: 0.140625
    jDirectionIncrementInDegrees: 0.09375

The UKV model X coordinate from MetOffice Datahub is as follows:

    DimCoord :  projection_x_coordinate / (m)
    points: [
        -575999.97097653, -573999.97097653, ...,  330000.02902347,
         332000.02902347]
    shape: (455,)
    dtype: float64
    standard_name: 'projection_x_coordinate'
    coord_system: LambertAzimuthalEqualArea(
        latitude_of_projection_origin=54.9,
        longitude_of_projection_origin=-2.5,
        false_easting=0.0,
        false_northing=0.0,
        ellipsoid=GeogCS(semi_major_axis=6378137.0, semi_minor_axis=6356752.314140356))

Which differs from the CEDA UKV data, as that is projected onto an OSGB grid:

    DimCoord :  projection_x_coordinate / (m)
    points: [-238000., -236000., ...,  854000.,  856000.]
    shape: (548,)
    dtype: float64
    standard_name: 'projection_x_coordinate'
    coord_system: TransverseMercator(
        latitude_of_projection_origin=49.0,
        longitude_of_central_meridian=-2.0,
        false_easting=400000.0,
        false_northing=-100000.0,
        scale_factor_at_central_meridian=0.0,
        ellipsoid=GeogCS(semi_major_axis=6377563.4, semi_minor_axis=6356256.91))

Similar for the Y values, for MetOffice:

    DimCoord :  projection_y_coordinate / (m)
    points: [
        -576000.00814487, -574000.00814487, ...,  697999.99185513,
         699999.99185513]
    shape: (639,)
    dtype: float64
    standard_name: 'projection_y_coordinate'
    coord_system: LambertAzimuthalEqualArea(
        latitude_of_projection_origin=54.9,
        longitude_of_projection_origin=-2.5,
        false_easting=0.0,
        false_northing=0.0,
        ellipsoid=GeogCS(semi_major_axis=6378137.0, semi_minor_axis=6356752.314140356))

And for CEDA:

    DimCoord :  projection_y_coordinate / (m)
    points: [1222000., 1220000., ..., -182000., -184000.]
    shape: (704,)
    dtype: float64
    standard_name: 'projection_y_coordinate'
    coord_system: TransverseMercator(
        latitude_of_projection_origin=49.0,
        longitude_of_central_meridian=-2.0,
        false_easting=400000.0,
        false_northing=-100000.0,
        scale_factor_at_central_meridian=0.0,
        ellipsoid=GeogCS(semi_major_axis=6377563.4, semi_minor_axis=6356256.91))


MetOffice's region selection allows you to choose a GB and EIRE region, which is then
defined as a bounding box from

    (-576719N, -576719E) to (700220N, 333408E)

Which doesn't quite line up with the values extracted from the grib above.


"""

import datetime as dt
import json
import logging
import os
import pathlib
import urllib.error
import urllib.request
from collections.abc import Callable, Iterator
from typing import TYPE_CHECKING, ClassVar, override

import xarray as xr
from joblib import delayed
from returns.result import Failure, ResultE, Success

from nwp_consumer.internal import entities, ports

if TYPE_CHECKING:
    import http.client

log = logging.getLogger("nwp-consumer")


class MetOfficeDatahubRawRepository(ports.RawRepository):
    """Repository implementation for data from MetOffice's DataHub service."""

    base_url: ClassVar[str] = (
        "https://data.hub.api.metoffice.gov.uk/atmospheric-models/1.0.0/orders"
    )

    request_url: str
    order_id: str
    dataspec: str
    _headers: dict[str, str]

    def __init__(self, order_id: str, api_key: str) -> None:
        """Create a new instance."""
        self._headers = {
            "Accept": "application/json",
            "apikey": api_key,
        }
        self.order_id = order_id
        self.request_url = f"{self.base_url}/{self.order_id}/latest"
        self.dataspec = os.getenv(
            "METOFFICE_DATASPEC",
            self.repository().optional_env["METOFFICE_DATASPEC"],
        )

    @staticmethod
    @override
    def repository() -> entities.RawRepositoryMetadata:

        return entities.RawRepositoryMetadata(
            name="MetOffice-Weather-Datahub",
            is_archive=False,
            is_order_based=True,
            max_connections=10,
            required_env=["METOFFICE_API_KEY", "METOFFICE_ORDER_ID"],
            optional_env={"METOFFICE_DATASPEC": "1.1.0"},
            postprocess_options=entities.PostProcessOptions(),
            available_models={
                "default": entities.Models.MO_UM_GLOBAL_10KM.with_region("india"),
                "um-global-10km-india": entities.Models.MO_UM_GLOBAL_10KM.with_region("india"),
                "um-global-10km-uk": entities.Models.MO_UM_GLOBAL_10KM.with_region("uk"),
                "um-global-10km-nl": entities.Models.MO_UM_GLOBAL_10KM\
                    .with_region("nl")\
                    .with_extra_parameters([entities.Parameter.PRESSURE_MSL]),
                "um-ukv-2km": entities.Models.MO_UM_UKV_2KM_LAEA,
            },
        )

    @staticmethod
    @override
    def model() -> entities.ModelMetadata:
        requested_model: str = os.getenv("MODEL", default="default")
        if requested_model not in MetOfficeDatahubRawRepository.repository().available_models:
            log.warn(
                f"Unknown model '{requested_model}' requested, falling back to default. "
                "MetOffice Datahub repository only supports "
                f"'{list(MetOfficeDatahubRawRepository.repository().available_models.keys())}'. "
                "Ensure MODEL environment variable is set to a valid model name.",
            )
            requested_model = "default"
        return MetOfficeDatahubRawRepository.repository().available_models[requested_model]

    @classmethod
    @override
    def authenticate(cls) -> ResultE["MetOfficeDatahubRawRepository"]:
        """Authenticate with the MetOffice DataHub service."""
        missing_envs = cls.repository().missing_required_envs()
        if len(missing_envs) > 0:
            return Failure(
                OSError(
                    f"Cannot authenticate with MetOffice DataHub service due to "
                    f"missing required environment variables: {', '.join(missing_envs)}",
                ),
            )
        api_key: str = os.environ["METOFFICE_API_KEY"]
        order_id: str = os.environ["METOFFICE_ORDER_ID"]
        return Success(cls(order_id=order_id, api_key=api_key))

    @override
    def fetch_init_data(
        self,
        it: dt.datetime,
    ) -> Iterator[Callable[..., ResultE[list[xr.DataArray]]]]:
        req: urllib.request.Request = urllib.request.Request(  # noqa: S310
            url=self.request_url + \
                f"?detail=MINIMAL&runfilter={it:%Y%m%d%H}&dataSpec={self.dataspec}",
            headers=self._headers,
            method="GET",
        )
        log.debug(
            f"Calling MetOffice Datahub at '{req.get_full_url()}'",
        )

        # Request the list of files
        try:
            response: http.client.HTTPResponse = urllib.request.urlopen(req, timeout=30)  # noqa: S310
        except Exception as e:
            yield delayed(Failure)(
                OSError(
                    "Unable to list files from MetOffice DataHub for order "
                    f"{self.order_id} at '{self.request_url}'. "
                    f"Ensure API key and Order ID are correct. Error context: {e}",
                ),
            )
            return
        try:
            data = json.loads(
                response.read().decode(response.info().get_param("charset") or "utf-8"),  # type: ignore
            )
        except Exception as e:
            yield delayed(Failure)(
                ValueError(
                    "Unable to decode JSON response from MetOffice DataHub. "
                    "Check the response from the '/latest' endpoint looks as expected. "
                    f"Error context: {e}",
                ),
            )
            return
        urls: list[str] = []
        if "orderDetails" in data and "files" in data["orderDetails"]:
            for filedata in data["orderDetails"]["files"]:
                if "fileId" in filedata and "+" not in filedata["fileId"]:
                    urls.append(
                        f"{self.request_url}/{filedata["fileId"]}/data?dataSpec={self.dataspec}",
                    )

        log.debug(
            f"Found {len(urls)} file(s) for init time '{it.strftime('%Y-%m-%d %H:%M')}' "
            f"in order '{self.order_id}'.",
        )

        for url in urls:
            yield delayed(self._download_and_convert)(url)

    def _download_and_convert(self, url: str) -> ResultE[list[xr.DataArray]]:
        """Download and convert a grib file from MetOffice Weather Datahub API.

        Args:
            url: The URL of the file of interest.
        """
        if "um-global" in self.model().name:
            return self._download(url).bind(self._convert_global)
        elif "um-ukv" in self.model().name:
            return self._download(url).bind(self._convert_ukv)
        else:
            return Failure(
                ValueError(
                    f"Unknown model '{self.model().name}' requested. "
                    "Ensure MODEL environment variable is set to a valid model name.",
                ),
            )

    def _download(self, url: str) -> ResultE[pathlib.Path]:
        """Download a grib file from MetOffice Weather Datahub API.

        Args:
            url: The URL of the file of interest.
        """
        local_path: pathlib.Path = (
            pathlib.Path(
                os.getenv(
                    "RAWDIR",
                    f"~/.local/cache/nwp/{self.repository().name}/{self.model().name}/raw",
                ),
            )
            / f"{url.split("/")[-2]}.grib"
        ).expanduser()

        # Only download the file if not already present
        if not local_path.exists() or local_path.stat().st_size == 0:
            local_path.parent.mkdir(parents=True, exist_ok=True)
            log.debug("Requesting file from MetOffice Weather Datahub API at: '%s'", url)

            req: urllib.request.Request = urllib.request.Request(  # noqa: S310
                url=url,
                headers=self._headers | {"Accept": "application/x-grib"},
                method="GET",
            )

            # Request the file
            try:
                response: http.client.HTTPResponse = urllib.request.urlopen(  # noqa: S310
                    req,
                    timeout=60,
                )
            except Exception as e:
                return Failure(
                    OSError(
                        "Unable to request file data from MetOffice DataHub at " f"'{url}': {e}",
                    ),
                )

            # Download the file
            log.debug("Downloading %s to %s", url, local_path)
            try:
                with local_path.open("wb") as f:
                    for chunk in iter(lambda: response.read(16 * 1024), b""):
                        f.write(chunk)
                        f.flush()
                    log.debug(
                        f"Downloaded '{url}' to '{local_path}' (%s bytes)",
                        local_path.stat().st_size,
                    )
            except Exception as e:
                return Failure(
                    OSError(
                        f"Error saving '{url}' to '{local_path}': {e}",
                    ),
                )

        return Success(local_path)

    @staticmethod
    def _convert_global(path: pathlib.Path) -> ResultE[list[xr.DataArray]]:
        """Convert a local grib file to xarray DataArrays.

        Args:
            path: The path to the file to convert.
        """
        try:
            # Read the file as a dataset, also reading the values of the keys in 'read_keys'
            ds: xr.Dataset = xr.open_dataset(
                path,
                engine="cfgrib",
                backend_kwargs={"read_keys": ["name", "parameterNumber"], "indexpath": ""},
                chunks={"time": 1, "step": -1},
            )
        except Exception as e:
            return Failure(OSError(f"Error opening '{path}' as xarray Dataset: {e}"))

        # Some parameters are surfaced in the dataset as 'unknown'
        # and have to be differentiated via the parameterNumber attribute
        # which lines up with the last number in the GRIB2 code specified below
        # https://datahub.metoffice.gov.uk/docs/glossary?sortOrder=GRIB2_CODE
        name = next(iter(ds.data_vars))
        parameter_number = ds[name].attrs["GRIB_parameterNumber"]
        match name, parameter_number:
            case "unknown", 192:
                ds = ds.rename({name: "u10"})
            case "unknown", 193:
                ds = ds.rename({name: "v10"})
            case "unknown", 194:
                ds = ds.rename({name: "wdir"})
            case "unknown", 195:
                ds = ds.rename({name: "si10"})
            case "unknown", 1:
                ds = ds.rename({name: "tcc"})
            case "unknown", _:
                log.warning(
                    f"Encountered unknown parameter with parameterNumber {parameter_number} "
                    f"in file '{path}'.",
                )

        try:
            ds = (
                ds.pipe(
                    entities.Parameter.rename_else_drop_ds_vars,
                    allowed_parameters=MetOfficeDatahubRawRepository.model().expected_coordinates.variable,
                )
                .rename(name_dict={"time": "init_time"})
                .expand_dims(dim="init_time")
            )

            if "step" not in ds.dims:
                ds = ds.expand_dims(dim="step")

            da: xr.DataArray = ds.to_dataarray(name=MetOfficeDatahubRawRepository.model().name)
            da = da.drop_vars(
                names=[
                    c
                    for c in ds.coords
                    if c not in MetOfficeDatahubRawRepository.model().expected_coordinates.dims
                ],
                errors="ignore",
            ).transpose(*MetOfficeDatahubRawRepository.model().expected_coordinates.dims)

            da = da.sortby(variables=["step", "variable", "longitude"])
            da = da.sortby(variables="latitude", ascending=False)

        except Exception as e:
            return Failure(
                ValueError(f"Error processing DataArray for path '{path}'. Error context: {e}"),
            )

        return Success([da])

    @staticmethod
    def _convert_ukv(path: pathlib.Path) -> ResultE[list[xr.DataArray]]:
        """Convert a local UKV grib file to xarray DataArrays.

        Args:
            path: The path to the file to convert.
        """
        try:
            # Read the file as a dataset, also reading the values of the keys in 'read_keys'
            ds: xr.Dataset = xr.open_dataset(
                path,
                engine="cfgrib",
                backend_kwargs={"read_keys": ["name", "parameterNumber"], "indexpath": ""},
                chunks={"time": 1, "step": -1},
            )
        except Exception as e:
            return Failure(OSError(f"Error opening '{path}' as xarray Dataset: {e}"))

        try:
            # Some parameters are surfaced in the dataset as 'unknown'
            # and have to be differentiated via the parameterNumber attribute
            # which lines up with the last number in the GRIB2 code specified below
            # https://datahub.metoffice.gov.uk/docs/glossary?sortOrder=GRIB2_CODE
            name = next(iter(ds.data_vars))
            parameter_number = ds[name].attrs["GRIB_parameterNumber"]
            match name, parameter_number:
                case "unknown", 192:
                    ds = ds.rename({name: "u10"})
                case "unknown", 193:
                    ds = ds.rename({name: "v10"})
                case "unknown", 194:
                    ds = ds.rename({name: "wdir"})
                case "unknown", 195:
                    ds = ds.rename({name: "si10"})
                case "unknown", 1:
                    ds = ds.rename({name: "tcc"})
                case "unknown", _:
                    log.warning(
                        f"Encountered unknown parameter with parameterNumber {parameter_number} "
                        f"in file '{path}'.",
                    )

            # Replace the coords with the expected values, as cfrib struggles to read them in.
            # The actual values have been determined by using iris.
            expected_coords = MetOfficeDatahubRawRepository.model().expected_coordinates
            if ds.sizes["x"] != len(expected_coords.x_laea):  # type: ignore
                return Failure(
                    ValueError(
                        f"Coordinate length of '{ds.sizes['x']}' for the x dimension for file "
                        f"'{path!s}' does not match expected length {len(expected_coords.x_laea)}",  # type: ignore
                    ),
                )
            if ds.sizes["y"] != len(expected_coords.y_laea):  # type: ignore
                return Failure(
                    ValueError(
                        f"Coordinate length of '{ds.sizes['y']}' for the y dimension for file "
                        f"'{path!s}' does not match expected length {len(expected_coords.y_laea)}",  # type: ignore
                    ),
                )
            # Assign coordinates to the dataset, and reverse y so it is descending, before
            # replacing the stand in values with the actual ones
            ds = (
                ds.assign_coords(
                    x=list(range(ds.sizes["x"])),
                    y=list(range(ds.sizes["y"])),
                )
                .sortby("y", ascending=False)
                .sortby("x")
                .assign_coords(
                    x=expected_coords.x_laea,
                    y=expected_coords.y_laea,
                )
                .rename({"x": "x_laea", "y": "y_laea"})
            )

            # Remove unwanted variables
            ds = (
                ds.pipe(
                    entities.Parameter.rename_else_drop_ds_vars,
                    allowed_parameters=MetOfficeDatahubRawRepository.model().expected_coordinates.variable,
                )
                .rename(name_dict={"time": "init_time"})
                .expand_dims(dim="init_time")
            )

            if "step" not in ds.dims:
                ds = ds.expand_dims(dim="step")

            da: xr.DataArray = ds.to_dataarray(name=MetOfficeDatahubRawRepository.model().name)
            da = (
                da.drop_vars(
                    names=[
                        c
                        for c in ds.coords
                        if c not in MetOfficeDatahubRawRepository.model().expected_coordinates.dims
                    ],
                    errors="ignore",
                )
                .transpose(*MetOfficeDatahubRawRepository.model().expected_coordinates.dims)
                .sortby(variables=["step", "variable", "x_laea"])
            )

        except Exception as e:
            return Failure(
                ValueError(
                    f"Error processing DataArray for path '{path}'. Error context: {e}",
                ),
            )

        return Success([da])



"""Interfaces for core services implementations.

The services module defines abstract interfaces that specify the signatures
any services implementations must obey in order to interact with the core.
"""

from .consumer_service import ConsumerService

__all__ = [
    "ConsumerService",
]



"""Model repository implementation for MetOffice Global data from CEDA.

Repository information:
=======================

The original model is from the UK Met Office, who don't provide their own archive.
CEDA (Centre for Environmental Data Analysis) host the data on their DAP server [2].
The CEDA catalogue for the Met Office Global can be found
`here <https://catalogue.ceda.ac.uk/uuid/86df725b793b4b4cb0ca0646686bd783>`_,
and the spec sheet from the Met Office is detailed in
`this PDF <https://www.metoffice.gov.uk/binaries/content/assets/metofficegovuk/pdf/data/global-atmospheric-model-17-km-resolution.pdf>`_.

For further details on the repository, see the
`CEDAFTPRawRepository.repository` implementation.

Data discrepancies and corrections
==================================

*MetOffice global* model data is stored on CEDA in segments:

- 4 areas for the northern hemisphere (A,B,C,D)
- 4 areas for the southern hemisphere (E,F,G,H)

Each area contains a subset of the data for a given time step.

Documented structure
--------------------

According to the MetOffice documentation [2], the files have the following structure::

    Northern hemisphere:
    - AreaA: Lat: 89.9 -> 0.3, Lon: -45 -> 45
    - AreaB: Lat: 89.9 -> 0.3, Lon: 45 -> 135
    - AreaC: Lat: 89.9 -> 0.3, Lon: 135 -> -135 (wraps around 180)
    - AreaD: Lat: 89.9 -> 0.3, Lon: -135 -> -45

    Southern hemisphere:
    - AreaE: Lat: -0.3 -> -89.9, Lon: -45 -> 45
    - AreaF: Lat: -0.3 -> -89.9, Lon: 45 -> 135
    - AreaG: Lat: -0.3 -> -89.9, Lon: 135 -> -135 (wraps around 180)
    - AreaH: Lat: -0.3 -> -89.9, Lon: -135 -> -45

With steps of 0.153 degrees in latitude and 0.234 degrees in longitude.

Actual structure
----------------

In my experience however, the data is not quite as described in the documentation.
Using the eccodes grib tool as shown::

    $ grib_ls -n geography -wcount=13 file.grib

I found that the grids are in fact as follows::

    - AreaA: Lat: 0 -> 89.856, Lon: 315 -> 45.09
    - AreaB: Lat: 0 -> 89.856, Lon: 45 -> 135.09
    - AreaC: Lat: 0 -> 89.856, Lon: 135 -> 225.09 (wraps around 180)
    - AreaD: Lat: 0 -> 89.856, Lon: 225 -> 315.09
    - AreaE: Lat: -89.856 -> 0, Lon: 315 -> 45.09
    - AreaF: Lat: -89.856 -> 0, Lon: 45 -> 135.09
    - AreaG: Lat: -89.856 -> 0, Lon: 135 -> 225.09 (wraps around 180)
    - AreaH: Lat: -89.856 -> 0, Lon: 225 -> 315.09

With steps of 0.156 degrees in latitude and 0.234 degrees in longitude.

.. important:: Key takeaways from this are:

    - The latitude values are in reverse order as described in the documentation
    - The longitude values overlap each other and combine to form a non-uniform step size
    - The step size is slightly different
    - Smaller lat/lon chunks are needed to allow for the partial area files to be written
      in parallel

As a result, the incoming data is modified to alleviate these issues.

"""

import base64
import datetime as dt
import itertools
import json
import logging
import os
import pathlib
import urllib.parse
import urllib.request
from collections.abc import Callable, Iterator
from typing import override

import cfgrib
import numpy as np
import xarray as xr
from joblib import delayed
from returns.result import Failure, ResultE, Success

from nwp_consumer.internal import entities, ports

log = logging.getLogger("nwp-consumer")


class CEDARawRepository(ports.RawRepository):
    """Repository implementation for the MetOffice global model data."""

    url_base: str = "https://dap.ceda.ac.uk/badc/ukmo-nwp/data"
    """The base URL for the CEDA DAP server."""
    _token: str
    """The authentication token for the CEDA DAP server."""

    def __init__(self, token: str) -> None:
        """Create a new instance."""
        self._token = token

    @staticmethod
    @override
    def repository() -> entities.RawRepositoryMetadata:
        return entities.RawRepositoryMetadata(
            name="CEDA",
            is_archive=True,
            is_order_based=False,
            max_connections=20,
            required_env=["CEDA_USER", "CEDA_PASS"],
            optional_env={},
            postprocess_options=entities.PostProcessOptions(),
            available_models={
                "default": entities.Models.MO_UM_GLOBAL_17KM.with_chunk_count_overrides(
                    {
                        "latitude": 8,
                        "longitude": 8,
                    },
                ).with_running_hours([0, 12]),  # 6 and 18 exist, but are lacking variables
                "mo-um-global": entities.Models.MO_UM_GLOBAL_17KM.with_chunk_count_overrides(
                    {
                        "latitude": 8,
                        "longitude": 8,
                    },
                ).with_running_hours([0, 12]),
                "mo-um-ukv": entities.Models.MO_UM_UKV_2KM_OSGB,
            },
        )

    @staticmethod
    @override
    def model() -> entities.ModelMetadata:
        requested_model: str = os.getenv("MODEL", default="default")
        if requested_model not in CEDARawRepository.repository().available_models:
            log.warn(
                f"Unknown model '{requested_model}' requested, falling back to default ",
                "CEDA repository only supports "
                f"'{list(CEDARawRepository.repository().available_models.keys())}'. "
                "Ensure MODEL environment variable is set to a valid model name.",
            )
            requested_model = "default"
        return CEDARawRepository.repository().available_models[requested_model]

    @override
    def fetch_init_data(
        self, it: dt.datetime,
    ) -> Iterator[Callable[..., ResultE[list[xr.DataArray]]]]:
        parameter_stubs: list[str] = [
            "Total_Downward_Surface_SW_Flux",
            "high_cloud_amount",
            "low_cloud_amount",
            "medium_cloud_amount",
            "relative_humidity_1_5m",
            "snow_depth",
            "temperature_1_5m",
            # "total_cloud",
            # "total_precipitation_rate", Exists, but only has 3 hourly steps
            "visibility_1_5m",
            "wind_u_10m",
            "wind_v_10m",
        ]

        file_stubs: list[str] = [
            "Wholesale1.grib",
            "Wholesale1T54.grib",
            "Wholesale2.grib",
            "Wholesale2T54.grib",
        ]

        if self.model().name == entities.Models.MO_UM_UKV_2KM_OSGB.name:
            for file in file_stubs:
                url = (
                    f"{self.url_base}/ukv-grib/{it:%Y/%m/%d}"
                    + f"/{it:%Y%m%d%H%M}_u1096_ng_umqv_{file}"
                )
                yield delayed(self._download_and_convert)(url=url)
        else:
            for param, area in itertools.product(parameter_stubs, [f"Area{c}" for c in "ABCDEFGH"]):
                url = (
                    f"{self.url_base}/global-grib/{it:%Y/%m/%d}"
                    + f"/{it:%Y%m%d%H}_WSGlobal17km_{param}_{area}_000144.grib"
                )
                yield delayed(self._download_and_convert)(url=url)

        pass

    def _download_and_convert(self, url: str) -> ResultE[list[xr.DataArray]]:
        """Download and convert a file to xarray DataArrays.

        Args:
            url: The URL of the file to download.
        """
        if "global-grib" in url:
            return self._download(url).bind(self._convert_global)
        elif "ukv-grib" in url:
            return self._download(url).bind(self._convert_ukv)
        else:
            return Failure(ValueError(f"Unknown URL type: {url}"))

    @classmethod
    @override
    def authenticate(cls) -> ResultE["CEDARawRepository"]:
        """Authenticate with CEDA.

        Returns:
            A Result containing the instantiated class if successful, or an error if not.
        """
        missing_envs = cls.repository().missing_required_envs()
        if len(missing_envs) > 0:
            return Failure(
                OSError(
                    f"Cannot authenticate with CEDA FTP service due to "
                    f"missing required environment variables: {', '.join(missing_envs)}",
                ),
            )
        username: str = os.environ["CEDA_USER"]
        password: str = os.environ["CEDA_PASS"]
        token: str = base64.b64encode(f"{username}:{password}".encode()).decode("ascii")
        request: urllib.request.Request = urllib.request.Request(
            method="POST",
            url="https://services-beta.ceda.ac.uk/api/token/create/",
            headers={"Authorization": f"Basic {token}"},
        )
        with urllib.request.urlopen(request, timeout=20) as response:  # noqa: S310
            if response.status != 200:
                return Failure(
                    OSError(
                        f"Failed to authenticate with CEDA: {response.status} {response.reason}",
                    ),
                )
            try:
                access_token: str = json.loads(response.read())["access_token"]
            except Exception as e:
                return Failure(
                    OSError(
                        f"Failed to parse CEDA access token: {e}",
                    ),
                )
        log.debug("Generated access token for CEDA")

        return Success(cls(token=access_token))

    def _download(self, url: str) -> ResultE[pathlib.Path]:
        """Download a file from CEDA.

        Args:
            url: The URL of the file to download.
        """
        local_path: pathlib.Path = (
            pathlib.Path(
                os.getenv(
                    "RAWDIR",
                    f"~/.local/cache/nwp/{self.repository().name}/{self.model().name}/raw",
                ),
            )
            / url.split("/")[-1]
        ).expanduser()

        # Don't download the file if it already exists
        if not local_path.exists():
            local_path.parent.mkdir(parents=True, exist_ok=True)
            log.debug("Sending request to CEDA DAP server for: '%s'", url)
            try:
                request: urllib.request.Request = urllib.request.Request(  # noqa: S310
                    method="GET",
                    url=url,
                    headers={"Authorization": f"Bearer {self._token}"},
                )
                response = urllib.request.urlopen(request, timeout=30)  # noqa: S310
            except Exception as e:
                return Failure(OSError(f"Error fetching {url}: {e}"))

            local_path.parent.mkdir(parents=True, exist_ok=True)
            log.debug("Downloading %s to %s", url, local_path)
            try:
                with local_path.open("wb") as f:
                    for chunk in iter(lambda: response.read(16 * 1024), b""):
                        f.write(chunk)
                        f.flush()
                log.debug(
                    f"Downloaded '{url}' to '{local_path}' (%s bytes)",
                    local_path.stat().st_size,
                )
            except Exception as e:
                return Failure(
                    OSError(
                        f"Error saving '{url}' to '{local_path}': {e}",
                    ),
                )

        return Success(local_path)

    @staticmethod
    def _convert_global(path: pathlib.Path) -> ResultE[list[xr.DataArray]]:
        """Convert a local grib file to xarray DataArrays.

        Args:
            path: The path to the file to convert.
        """
        try:
            dss: list[xr.Dataset] = cfgrib.open_datasets(path)
        except Exception as e:
            return Failure(
                OSError(
                    f"Error opening '{path}' as xarray Datasets: {e}",
                ),
            )
        try:
            processed_das: list[xr.DataArray] = []
            for ds in dss:
                ds = entities.Parameter.rename_else_drop_ds_vars(
                    ds=ds,
                    allowed_parameters=CEDARawRepository.model().expected_coordinates.variable,
                )
                # Ignore datasets with no variables of interest
                if len(ds.data_vars) == 0:
                    return Failure(
                        OSError(
                            f"No relevant variables found in '{path}'. "
                            "Ensure file contains the expected variables, "
                            "and that desired variables are not being dropped.",
                        ),
                    )
                da: xr.DataArray = (
                    ds.where(
                        ds.step
                        <= np.timedelta64(
                            CEDARawRepository.model().expected_coordinates.step[-1],
                            "h",
                        ),
                        drop=True,
                    )
                    .drop_vars(
                        names=[
                            c
                            for c in ds.coords
                            if c not in ["time", "step", "latitude", "longitude"]
                        ],
                    )
                    .rename(name_dict={"time": "init_time"})
                    .expand_dims(dim="init_time")
                    .to_dataarray(name=CEDARawRepository.model().name)
                )
                da = da.transpose(*CEDARawRepository.model().expected_coordinates.dims)
                if "longitude" in da.coords:
                    # We are dealing with the "AreaX" files:
                    # * remove the last value of the longitude dimension
                    #   as it overlaps with the next file
                    # * reverse the latitude dimension to be in descending order
                    da = da.isel(longitude=slice(None, -1), latitude=slice(None, None, -1))
                # Put each variable into its own DataArray:
                # * Each raw file does not contain a full set of parameters
                # * and so may not produce a contiguous subset of the expected coordinates.
                processed_das.extend(
                    [
                        da.where(cond=da.coords["variable"] == v, drop=True)
                        for v in da.coords["variable"].values
                    ],
                )
        except Exception as e:
            return Failure(
                ValueError(
                    f"Error processing {path} to DataArray: {e}",
                ),
            )
        return Success(processed_das)

    @staticmethod
    def _convert_ukv(path: pathlib.Path) -> ResultE[list[xr.DataArray]]:
        """Convert a local wholesale grib file to xarray DataArrays."""
        # Load the wholesale file as a list of datasets
        # * cfgrib loads multiple hypercubes for a single multi-parameter grib file
        # * Can also set backend_kwargs={"indexpath": ""}, to avoid the index file
        try:
            dss: list[xr.Dataset] = cfgrib.open_datasets(
                path=path.as_posix(),
                chunks={"time": 1, "step": -1, "variable": -1, "x": "auto", "y": "auto"},
                backend_kwargs={"indexpath": ""},
            )
        except Exception as e:
            return Failure(
                OSError(
                    f"Error opening '{path}' as xarray Datasets: {e}",
                ),
            )

        processed_das: list[xr.DataArray] = []
        try:
            for ds in dss:
                # Ensure the temperature is defined at 1 meter above ground level
                # * In the early NWPs (definitely in the 2016-03-22 NWPs):
                #   - `heightAboveGround` only has one entry ("1" meter above ground)
                #   - `heightAboveGround` isn't set as a dimension for `t`.
                # * In later NWPs, 'heightAboveGround' has 2 values (0, 1) and is a
                #   dimension for `t`.
                if "t" in ds and "heightAboveGround" in ds["t"].dims:
                    ds = ds.sel(heightAboveGround=1)

                # Delete unnecessary data variables
                ds = entities.Parameter.rename_else_drop_ds_vars(
                    ds=ds,
                    allowed_parameters=CEDARawRepository.model().expected_coordinates.variable,
                )
                if len(ds.data_vars) == 0:
                    continue

                ds = (
                    ds.rename({"time": "init_time"})
                    .expand_dims("init_time")
                    .drop_vars(
                        names=[
                            c
                            for c in ds.coords
                            if c not in CEDARawRepository.model().expected_coordinates.dims
                        ],
                        errors="ignore",
                    )
                    .where(
                        ds.step
                        <= np.timedelta64(
                            CEDARawRepository.model().expected_coordinates.step[-1],
                            "h",
                        ),
                        drop=True,
                    )
                )
                # Adapted from https://stackoverflow.com/a/62667154 and
                # https://github.com/SciTools/iris-grib/issues/140#issuecomment-1398634288
                northing: list[int] = CEDARawRepository.model().expected_coordinates.y_osgb  # type: ignore
                easting: list[int] = CEDARawRepository.model().expected_coordinates.x_osgb  # type: ignore
                if ds.sizes["values"] != len(northing) * len(easting):
                    raise ValueError(
                        f"dataset has {ds.sizes['values']} values, "
                        f"but expected {len(northing) * len(easting)}",
                    )
                ds = ds.assign_coords(
                    {
                        "x_osgb": ("values", np.tile(easting, reps=len(northing))),
                        "y_osgb": ("values", np.repeat(northing, repeats=len(easting))),
                    },
                )
                # Set `values` to be a MultiIndex, indexed by `y` and `x`, then unstack
                # * This gets rid of the `values` dimension and indexes
                #   the data variables using `y` and `x`.
                ds = ds.set_index(values=("y_osgb", "x_osgb")).unstack("values")
                da: xr.DataArray = (
                    ds.to_array(name=CEDARawRepository.model().name)
                    .sortby(variables=["step", "variable", "x_osgb"])
                    .sortby(variables="y_osgb", ascending=False)
                    .transpose("init_time", "step", "variable", "y_osgb", "x_osgb")
                    .drop_vars(
                        names=[
                            c
                            for c in ds.coords
                            if c not in CEDARawRepository.model().expected_coordinates.dims
                        ],
                        errors="ignore",
                    )
                )
                # Put each variable into its own DataArray:
                # * Each raw file does not contain a full set of parameters
                # * and so may not produce a contiguous subset of the expected coordinates.
                processed_das.extend(
                    [
                        da.where(cond=da.coords["variable"] == v, drop=True)
                        for v in da.coords["variable"].values
                    ],
                )
        except Exception as e:
            return Failure(
                ValueError(
                    f"Error processing {path} to DataArray: {e}",
                ),
            )

        return Success(processed_das)



"""Stdout notification repository implementation."""

import logging
from typing import override

from returns.result import ResultE, Success

from nwp_consumer.internal import entities, ports

log = logging.getLogger("nwp-consumer")


class StdoutNotificationRepository(ports.NotificationRepository):
    """Stdout notification repository."""

    @override
    def notify(
        self,
        message: entities.StoreCreatedNotification | entities.StoreAppendedNotification,
    ) -> ResultE[str]:
        log.info(f"{message}")
        return Success("Notification sent to stdout successfully.")



"""Repository implementation for NOAA GFS data stored in S3.

This module contains the implementation of the model repository for the
NOAA GFS data stored in an S3 bucket.

Repository Information
======================

TODO: provide links etc

Documented Structure
--------------------

The data is provided with a file per step. There is a difference between what is
provided in the 0-step file and the remaning 3-hourly files. The 0-step file
does not contain e.g. downward shortwave radiation flux and downward longwave
radiation flux. As such, these will be NaNs in the 0-step file. This is because
the 0-step file is contains an analysis of the forecast, while the remaining
files are the forecast itself. I think this is just a quirk of NOAA's distribution.

See Also:
    - https://www.nco.ncep.noaa.gov/pmb/products/gfs/
    - https://www.nco.ncep.noaa.gov/pmb/products/gfs/gfs.t00z.pgrb2.1p00.f000.shtml
    - https://www.nco.ncep.noaa.gov/pmb/products/gfs/gfs.t00z.pgrb2.1p00.f003.shtml

TODO: document filestructure
"""

import datetime as dt
import logging
import os
import pathlib
import re
from collections.abc import Callable, Iterator
from typing import override

import cfgrib
import s3fs
import xarray as xr
from joblib import delayed
from returns.result import Failure, ResultE, Success

from nwp_consumer.internal import entities, ports

log = logging.getLogger("nwp-consumer")


class NOAAS3RawRepository(ports.RawRepository):
    """Model repository implementation for GFS data stored in S3."""

    @staticmethod
    @override
    def repository() -> entities.RawRepositoryMetadata:
        return entities.RawRepositoryMetadata(
            name="NOAA-GFS-S3",
            is_archive=False,
            is_order_based=False,
            max_connections=100,
            required_env=[],
            optional_env={},
            postprocess_options=entities.PostProcessOptions(),
            available_models={
                "default": entities.Models.NCEP_GFS_1DEGREE,
                "gfs-1deg": entities.Models.NCEP_GFS_1DEGREE,
            },
        )

    @staticmethod
    @override
    def model() -> entities.ModelMetadata:
        return NOAAS3RawRepository.repository().available_models["default"]

    @override
    def fetch_init_data(
        self,
        it: dt.datetime,
    ) -> Iterator[Callable[..., ResultE[list[xr.DataArray]]]]:
        # List relevant files in the s3 bucket
        bucket_path: str = f"noaa-gfs-bdp-pds/gfs.{it:%Y%m%d}/{it:%H}/atmos"
        try:
            fs = s3fs.S3FileSystem(anon=True)
            urls: list[str] = [
                f"s3://{f}"
                for f in fs.ls(bucket_path)
                if self._wanted_file(
                    filename=f.split("/")[-1],
                    it=it,
                    steps=self.model().expected_coordinates.step,
                )
            ]
        except Exception as e:
            yield delayed(Failure)(
                ValueError(
                    f"Failed to list file in bucket path '{bucket_path}'. "
                    "Ensure the path exists and the bucket does not require auth. "
                    f"Encountered error: '{e}'",
                ),
            )
            return

        if len(urls) == 0:
            yield delayed(Failure)(
                ValueError(
                    f"No files found for init time '{it:%Y-%m-%d %H:%M}'. "
                    "in bucket path '{bucket_path}'. Ensure files exists at the given path "
                    "with the expected filename pattern. ",
                ),
            )

        for url in urls:
            yield delayed(self._download_and_convert)(url=url, it=it)

    @classmethod
    @override
    def authenticate(cls) -> ResultE["NOAAS3RawRepository"]:
        return Success(cls())

    def _download_and_convert(self, url: str, it: dt.datetime) -> ResultE[list[xr.DataArray]]:
        """Download and convert a file from S3.

        Args:
            url: The URL to the S3 object.
            it: The init time of the object in question, used in the saved path
        """
        return self._download(url=url, it=it).bind(self._convert)

    def _download(self, url: str, it: dt.datetime) -> ResultE[pathlib.Path]:
        """Download a grib file from NOAA S3.

        The URLs have the following format::

          https://noaa-gfs-bdp-pds.s3.amazonaws.com/gfs.20230911/06/atmos/gfs.t06z.pgrb2.1p00.f087
          <------------------bucket---------------><---inittime--->      <-------filename----step>

        Args:
            url: The URL to the S3 object.
            it: The init time of the object in question, used in the saved path
        """
        local_path: pathlib.Path = (
            pathlib.Path(
                os.getenv(
                    "RAWDIR",
                    f"~/.local/cache/nwp/{self.repository().name}/{self.model().name}/raw",
                ),
            )
            / it.strftime("%Y/%m/%d/%H")
            / (url.split("/")[-1] + ".grib")
        ).expanduser()

        # Only download the file if not already present
        if local_path.exists():
            return Success(local_path)

        local_path.parent.mkdir(parents=True, exist_ok=True)
        log.debug("Requesting file from S3 at: '%s'", url)

        fs = s3fs.S3FileSystem(anon=True)
        try:
            if not fs.exists(url):
                raise FileNotFoundError(f"File not found at '{url}'")

            with local_path.open("wb") as lf, fs.open(url, "rb") as rf:
                for chunk in iter(lambda: rf.read(12 * 1024), b""):
                    lf.write(chunk)
                    lf.flush()

        except Exception as e:
            return Failure(
                OSError(
                    f"Failed to download file from S3 at '{url}'. Encountered error: {e}",
                ),
            )

        # For some reason, the GFS files are about 2MB larger when downloaded
        # then their losted size in AWS. I'd be interested to know why!
        if local_path.stat().st_size < fs.info(url)["size"]:
            return Failure(
                ValueError(
                    f"File size mismatch from file at '{url}': "
                    f"{local_path.stat().st_size} != {fs.info(url)['size']} (remote). "
                    "File may be corrupted.",
                ),
            )

        # Also download the associated index file
        # * This isn't critical, but speeds up reading the file in when converting
        # TODO: Re-incorporate this when https://github.com/ecmwf/cfgrib/issues/350
        # TODO: is resolved. Currently downloaded index files are ignored due to
        # TODO: path differences once downloaded.
        # index_url: str = url + ".idx"
        # index_path: pathlib.Path = local_path.with_suffix(".grib.idx")
        # try:
        #     with index_path.open("wb") as lf, fs.open(index_url, "rb") as rf:
        #         for chunk in iter(lambda: rf.read(12 * 1024), b""):
        #             lf.write(chunk)
        #             lf.flush()
        # except Exception as e:
        #     log.warning(
        #         f"Failed to download index file from S3 at '{url}'. "
        #         "This will require a manual indexing when converting the file. "
        #         f"Encountered error: {e}",
        #     )

        return Success(local_path)

    @staticmethod
    def _convert(path: pathlib.Path) -> ResultE[list[xr.DataArray]]:
        """Convert a GFS file to an xarray DataArray collection.

        Args:
            path: The path to the local grib file.
        """
        try:
            # Use some options when opening the datasets:
            # * 'filter_by_keys' reduces the number of variables loaded to only those
            #   with names of interest. "t" is filtered out as it exists in multiple
            #   levels
            filters: list[dict[str, list[str] | list[int] | int]] = [
                {
                    "cfVarName": ["tcc", "hcc", "lcc", "mcc"],
                    "level": 0,
                    "typeOfLevel": [
                        "highCloudLayer",
                        "lowCloudLayer",
                        "middleCloudLayer",
                        "convectiveCloudLayer",
                    ],
                    "stepType": ["instant"],
                },
                {"cfVarName": ["u10", "v10"]},
                {"cfVarName": ["u100", "v100"]},
                {
                    "typeOfLevel": ["surface", "heightAboveGround"],
                    "level": [0, 2],
                    "stepType": ["instant"],
                },
                {
                    "typeOfLevel": ["surface"],
                    "stepType": ["avg"],
                    "cfVarName": ["sdswrf", "sdlwrf"],
                },
            ]
            ds: xr.Dataset = xr.merge(
                [
                    cfgrib.open_dataset(path.as_posix(), backend_kwargs={"filter_by_keys": f})
                    for f in filters
                ],
                compat="minimal",
            ).drop_vars("t", errors="ignore")
        except Exception as e:
            return Failure(
                ValueError(
                    f"Error opening '{path}' as list of xarray Datasets: {e}",
                ),
            )

        if len(ds.data_vars) == 0:
            return Failure(
                ValueError(
                    f"No datasets found in '{path}'. File may be corrupted. "
                    "A redownload of the file may be required.",
                ),
            )

        try:
            ds = ds.drop_vars("sdwe", errors="ignore")  # Datasets contain both SDWE and SD
            ds = entities.Parameter.rename_else_drop_ds_vars(
                ds=ds,
                allowed_parameters=NOAAS3RawRepository.model().expected_coordinates.variable,
            )
            da: xr.DataArray = (
                ds.drop_vars(
                    names=[
                        c for c in ds.coords if c not in ["time", "step", "latitude", "longitude"]
                    ],
                )
                .rename(name_dict={"time": "init_time"})
                .expand_dims(dim="init_time")
                .expand_dims(dim="step")
                .to_dataarray(name=NOAAS3RawRepository.model().name)
            )
            da = (
                da.drop_vars(
                    names=[
                        c
                        for c in da.coords
                        if c not in NOAAS3RawRepository.model().expected_coordinates.dims
                    ],
                )
                .transpose(*NOAAS3RawRepository.model().expected_coordinates.dims)
                .assign_coords(coords={"longitude": (da.coords["longitude"] + 180) % 360 - 180})
                .sortby(variables=["step", "variable", "longitude"])
                .sortby(variables="latitude", ascending=False)
            )
        except Exception as e:
            return Failure(
                ValueError(
                    f"Error processing dataset from '{path}' to DataArray: {e}",
                ),
            )

        return Success([da])

    @staticmethod
    def _wanted_file(filename: str, it: dt.datetime, steps: list[int]) -> bool:
        """Determine if a file is wanted based on the init time and max step.

        See module docstring for file naming convention.
        """
        pattern: str = r"^gfs\.t(\d{2})z\.pgrb2\.1p00\.f(\d{3})$"
        match: re.Match[str] | None = re.search(pattern=pattern, string=filename)
        if match is None:
            return False
        if int(match.group(1)) != it.hour:
            return False
        return int(match.group(2)) in steps



"""Domain classes for repository metadata.

Sources of NWP data have attributes both pertaining to and apart from
the data they deliver. This module defines classes for metadata that
tracks relevant information about the model repository and the data
it provides. This might be helpful in determining the quality of the
data, defining pipelines for processing, or establishing the availability
for a live service.

In this instance, the `RawRepositoryMetadata` refers to information
about the repository where NWP data produced by the model resides.
"""

import dataclasses
import datetime as dt
import os

from .modelmetadata import ModelMetadata
from .postprocess import PostProcessOptions


@dataclasses.dataclass(slots=True)
class RawRepositoryMetadata:
    """Metadata for an NWP Model repository."""

    name: str
    """The name of the model repository."""

    is_archive: bool
    """Whether the repository is a complete archival set.

    Archival datasets are able to backfill old data.
    Non-archival datasets only provide a limited window of data.
    """

    is_order_based: bool
    """Whether the repository is order-based.

    This means parameters cannot be chosen freely,
    but rather are defined by pre-selected agreements with the provider.
    """

    required_env: list[str]
    """Environment variables required for usage."""

    optional_env: dict[str, str]
    """Optional environment variables."""

    max_connections: int
    """The maximum number of simultaneous connections allowed to the model repository.

    This determines the maximum level of concurrency that can be achieved when
    downloading data from the repository.
    """

    postprocess_options: PostProcessOptions
    """Options for post-processing the data."""

    available_models: dict[str, ModelMetadata]
    """A dictionary of available models and their metadata."""

    def determine_latest_it_from(
            self, t: dt.datetime,
            running_hours: list[int],
            delay_minutes:int) -> dt.datetime:
        """Determine the latest available initialization time from a given time.

        Args:
            t: The time from which to determine the latest initialization time.
            running_hours: A list of hours at which the model runs each day.
            delay_minutes: The delay in minutes after the initialization time
                before data is available.

        Returns:
            The latest available initialization time prior to the given time.
        """
        it = (
            t.replace(minute=0, second=0, microsecond=0) - dt.timedelta(minutes=delay_minutes)
        ).replace(minute=0)
        while it.hour not in running_hours:
            it -= dt.timedelta(hours=1)

        return it

    def missing_required_envs(self) -> list[str]:
        """Get a list of unset required environment variables.

        Returns:
            A list of missing environment variables.
        """
        return [var for var in self.required_env if var not in os.environ]

    def __str__(self) -> str:
        """Return a pretty-printed string representation of the metadata."""
        pretty: str = "".join(
            (
                "Model Repository: ",
                f"\n\t{self.name} ({'archive' if self.is_archive else 'live/rolling'} dataset.)",
                "\nEnvironment variables:",
                "\n\tRequired:",
                "\n".join(f"\t\t{var}" for var in self.required_env),
                "\n\tOptional:",
                "\n".join(f"\t\t{var}={val}" for var, val in self.optional_env.items()),
            ),
        )
        return pretty



import datetime as dt
from collections.abc import Callable, Iterator
from typing import override

import numpy as np
import xarray as xr
from joblib import delayed
from returns.result import ResultE, Success

from nwp_consumer.internal import entities, ports


class DummyRawRepository(ports.RawRepository):
    @classmethod
    @override
    def authenticate(cls) -> ResultE["DummyRawRepository"]:
        return Success(cls())

    @staticmethod
    @override
    def repository() -> entities.RawRepositoryMetadata:
        return entities.RawRepositoryMetadata(
            name="ACME-Test-Models",
            is_archive=False,
            is_order_based=False,
            max_connections=4,
            required_env=[],
            optional_env={},
            postprocess_options=entities.PostProcessOptions(),
            available_models={},
        )

    @staticmethod
    @override
    def model() -> entities.ModelMetadata:
        return entities.ModelMetadata(
            name="simple-random",
            resolution="17km",
            expected_coordinates=entities.NWPDimensionCoordinateMap(
                init_time=[dt.datetime(2021, 1, 1, 0, 0, tzinfo=dt.UTC)],
                step=list(range(0, 48, 1)),
                variable=[
                    entities.Parameter.TEMPERATURE_SL,
                    entities.Parameter.DOWNWARD_SHORTWAVE_RADIATION_FLUX_GL,
                    entities.Parameter.CLOUD_COVER_HIGH,
                ],
                latitude=np.linspace(90, -90, 721).tolist(),
                longitude=np.linspace(-180, 179.8, 1440).tolist(),
            ),
            running_hours=[0, 6, 12, 18],
            delay_minutes=60,
        )

    @override
    def fetch_init_data(
        self, it: dt.datetime,
    ) -> Iterator[Callable[..., ResultE[list[xr.DataArray]]]]:
        def gen_dataset(step: int, variable: str) -> ResultE[list[xr.DataArray]]:
            """Define a generator that provides one variable at one step."""
            da = xr.DataArray(
                name=self.model().name,
                dims=["init_time", "step", "variable", "latitude", "longitude"],
                data=np.random.rand(1, 1, 1, 721, 1440),
                coords=self.model().expected_coordinates.to_pandas()
                | {
                    "init_time": [np.datetime64(it.replace(tzinfo=None), "ns")],
                    "step": [step],
                    "variable": [variable],
                },
            )
            return Success([da])

        for s in self.model().expected_coordinates.step:
            for v in self.model().expected_coordinates.variable:
                yield delayed(gen_dataset)(s, v.value)


class DummyNotificationRepository(ports.NotificationRepository):
    @override
    def notify(
        self,
        message: entities.StoreAppendedNotification | entities.StoreCreatedNotification,
    ) -> ResultE[str]:
        return Success(str(message))



import datetime as dt
import shutil
import unittest

import xarray as xr
from returns.pipeline import is_successful

from nwp_consumer.internal.services.consumer_service import ConsumerService

from ._dummy_adaptors import DummyNotificationRepository, DummyRawRepository


class TestParallelConsumer(unittest.TestCase):
    def test_consume(self) -> None:
        """Test the consume method of the ParallelConsumer class."""

        test_consumer = ConsumerService.from_adaptors(
            model_adaptor=DummyRawRepository,
            notification_adaptor=DummyNotificationRepository,
        ).unwrap()

        result = test_consumer.consume(period=dt.datetime(2021, 1, 1, tzinfo=dt.UTC))

        self.assertTrue(is_successful(result), msg=result)

        da: xr.DataArray = xr.open_dataarray(result.unwrap(), engine="zarr")

        self.assertEqual(
            list(da.sizes.keys()),
            ["init_time", "step", "variable", "latitude", "longitude"],
        )

        path = result.unwrap()
        shutil.rmtree(path)


if __name__ == "__main__":
    unittest.main()
